from .io import opener
from pathlib import Path
import numpy as np
import logging
from math import ceil
from datetime import datetime
from io import BytesIO
import xarray
from typing import List, Any, Dict, Tuple, Sequence, Optional
from typing.io import TextIO


def rinexobs2(fn: Path,
              use: Sequence[str]=None,
              tlim: Tuple[datetime, datetime]=None,
              useindicators: bool=False,
              meas: Sequence[str]=None,
              verbose: bool=False) -> xarray.Dataset:

    if isinstance(use, str):
        use = [use]

    if use is None or not use[0].strip():
        use = ('C', 'E', 'G', 'J', 'R', 'S')

    obs = None
    for u in use:
        o = rinexsystem2(fn, u, tlim, useindicators, meas, verbose)
        if o is not None:
            if obs is None:
                attrs = o.attrs
                obs = o
            else:
                obs = xarray.merge((obs, o))

    if obs is not None:
        obs.attrs = attrs

    return obs


def rinexsystem2(fn: Path,
                 system: str,
                 tlim: Tuple[datetime, datetime]=None,
                 useindicators: bool=False,
                 meas: Sequence[str]=None,
                 verbose: bool=False) -> xarray.Dataset:
    """
    procss RINEX OBS data
    fn: RINEX OBS 2 filename
    system: 'G', 'R', or similar
    meas:  'L1C'  or  ['L1C', 'C1C'] or similar
    """
    Lf = 14
    assert isinstance(system, str)
# %% allocation
    Nsvmax = 32  # FIXME
    times = obstime2(fn)  # < 10 ms for 24 hour 15 second cadence
    hdr = obsheader2(fn, useindicators, meas)

    if hdr['systems'] != 'M' and system != hdr['systems']:
        return

    Nl_sv = ceil(hdr['Nobs']/5)

    Npages = hdr['Nobsused']*3 if useindicators else hdr['Nobsused']
    data = np.empty((Npages, times.size, Nsvmax))
    if data.size == 0:
        return

    data.fill(np.nan)


# %% start reading
    with opener(fn) as f:
        # skip header
        for ln in f:
            if 'END OF HEADER' in ln:
                break

# %% process data
        j = -1  # not enumerate in case of time error
        for ln in f:
            time_epoch = _timeobs(ln)
            if time_epoch is None:
                continue

            j += 1

            if tlim is not None:
                assert isinstance(tlim[0], datetime), 'time bounds are specified as datetime.datetime'
                if time_epoch < tlim[0]:
                    _skip(f, ln, hdr['Nobs'])
                    continue
                elif time_epoch > tlim[1]:
                    break
# %%
            eflag = int(ln[28])
            if eflag not in (0, 1, 5, 6):  # EPOCH FLAG
                logging.info(f'{time_epoch}: epoch flag {eflag}')
                continue

            if verbose:
                print(time_epoch, end="\r")

            try:
                toffset = ln[68:80]
            except ValueError:
                toffset = None
# %% get SV indices
            sv = _getsvind(f, ln)
# %% select one, a few, or all satellites
            iuse = [i for i, s in enumerate(sv) if s[0] == system]
            if len(iuse) == 0:
                _skip(f, ln, hdr['Nobs'], sv)
                continue

            gsv = np.array(sv)[iuse]
# %% assign data for each time step
            raws = ''
            for i, s in enumerate(sv):
                # don't process discarded satellites
                if not s[0] == system:
                    for _ in range(Nl_sv):
                        f.readline()
                    continue
                # .rstrip() necessary to handle variety of files and Windows vs. Unix
                raw = ''
                for _ in range(Nl_sv):
                    raw += f.readline()[:80].rstrip()

                raws += raw + '\n'
            """
            it is about 5x faster to call np.genfromtxt() for all sats and then index,
            vs. calling np.genfromtxt() for each sat.
            """
            # can't use "usecols" with "delimiter"
            darr = np.genfromtxt(BytesIO(raws.encode('ascii')), delimiter=[Lf, 1, 1]*hdr['Nobs'])
            darr = np.atleast_2d(darr)
            assert darr.shape[0] == len(gsv)

# %% select only "used" satellites
            for i, k in enumerate(hdr['fields_ind']):
                isv = [int(s[1:])-1 for s in gsv]
                if useindicators:
                    data[i*3, j, isv] = darr[:, k*3]
                    # FIXME which other should be excluded?
                    if hdr['fields'][k] not in ('S1', 'S2'):
                        if hdr['fields'][k] in ('L1', 'L2'):
                            data[i*3+1, j, isv] = darr[:, k*3+1]

                        data[i*3+2, j, isv] = darr[:, k*3+2]
                else:
                    data[i, j, isv] = darr[:, k]
# %% output gathering
    fields = []
    for field in hdr['fields']:
        fields.append(field)
        if useindicators:
            if field not in ('S1', 'S2'):
                if field in ('L1', 'L2'):
                    fields.append(f'{field}lli')
                else:
                    fields.append(None)
                fields.append(f'{field}ssi')
            else:
                fields.extend([None, None])

    if np.isnan(data).all():
        return

    obs = xarray.Dataset(coords={'time': times,
                                 'sv': [f'{system}{i:02d}' for i in range(1, Nsvmax+1)]})

    for i, k in enumerate(fields):
        # FIXME: for limited time span reads, this drops unused data variables
        # if np.isnan(data[i, ...]).all():
            # continue
        if k is None:
            continue
        obs[k] = (('time', 'sv'), data[i, :, :])

    obs = obs.dropna(dim='sv', how='all')
    obs = obs.dropna(dim='time', how='all')  # when tlim specified
# %% attributes
    obs.attrs['filename'] = fn.name
    obs.attrs['version'] = hdr['version']
    obs.attrs['position'] = hdr['position']
    obs.attrs['rinextype'] = 'obs'
    obs.attrs['toffset'] = toffset

    return obs


def obsheader2(f: TextIO,
               useindicators: bool=False,
               meas: Sequence[str]=None) -> Dict[str, Any]:

    if isinstance(f, Path):
        fn = f
        with opener(fn) as f:
            return obsheader2(f, useindicators, meas)

# %% selection
    if isinstance(meas, str):
        meas = [meas]

    if not meas or not meas[0].strip():
        meas = None

    header: Dict[str, Any] = {}
    Nobs = 0  # not None due to type checking

    for l in f:
        if "END OF HEADER" in l:
            break

        h = l[60:80].strip()
        c = l[:60]
# %% measurement types
        if '# / TYPES OF OBSERV' in h:
            if Nobs == 0:
                Nobs = int(c[:6])

            c = c[6:].split()  # NOT within "if Nobs"
# %%
        if h not in header:  # Header label
            header[h] = c  # string with info
        else:  # concatenate
            if isinstance(header[h], str):
                header[h] += " " + c
            elif isinstance(header[h], list):
                header[h] += c
            else:
                raise ValueError(f'not sure what {c} is')
# %% useful values
    header['version'] = float(header['RINEX VERSION / TYPE'][:9])  # %9.2f
    header['systems'] = header['RINEX VERSION / TYPE'][40]
    header['Nobs'] = Nobs
    # list with x,y,z cartesian
    header['position'] = [float(j) for j in header['APPROX POSITION XYZ'].split()]
# %% observation types
    header['fields'] = header['# / TYPES OF OBSERV']
    if Nobs != len(header['fields']):
        raise ValueError(f'{f.name} header read incorrectly')

    if isinstance(meas, (tuple, list, np.ndarray)):
        ind = np.zeros(len(header['fields']), dtype=bool)
        for m in meas:
            for i, f in enumerate(header['fields']):
                if f.startswith(m):
                    ind[i] = True

        # zeros, NOT empty!
        sysind = np.zeros(Nobs*3, dtype=bool)  # *3 due to LLI, SSI
        if useindicators:
            for j, i in enumerate(ind):
                sysind[j*3:j*3+3] = i
        else:
            sysind[::3] = ind

        header['fields_ind'] = np.nonzero(sysind)[0]
    else:
        ind = slice(None)
        header['fields_ind'] = np.arange(Nobs) if useindicators else np.arange(0, Nobs*3, 3)

    header['fields'] = np.array(header['fields'])[ind].tolist()

    if isinstance(header['fields_ind'], slice):
        Nobsused = header['Nobs']*3
    elif header['fields_ind'].dtype == bool:
        Nobsused = np.count_nonzero(header['fields_ind'])
    else:
        Nobsused = header['fields_ind'].size
    header['Nobsused'] = Nobsused
# %%
    if '# OF SATELLITES' in header:
        header['# OF SATELLITES'] = int(header['# OF SATELLITES'][:6])
# %% time
    t0s = header['TIME OF FIRST OBS']
    # NOTE: must do second=int(float()) due to non-conforming files
    header['t0'] = datetime(year=int(t0s[:6]), month=int(t0s[6:12]), day=int(t0s[12:18]),
                            hour=int(t0s[18:24]), minute=int(t0s[24:30]), second=int(float(t0s[30:36])),
                            microsecond=int(float(t0s[30:43]) % 1 * 1000000))

    try:
        t0s = header['TIME OF LAST OBS']
        # NOTE: must do second=int(float()) due to non-conforming files
        header['t1'] = datetime(year=int(t0s[:6]), month=int(t0s[6:12]), day=int(t0s[12:18]),
                                hour=int(t0s[18:24]), minute=int(t0s[24:30]), second=int(float(t0s[30:36])),
                                microsecond=int(float(t0s[30:43]) % 1 * 1000000))
    except KeyError:
        pass

    try:  # This key is OPTIONAL
        header['interval'] = float(header['INTERVAL'][:10])
    except KeyError:
        header['interval'] = None

    return header


def _getsvind(f: TextIO, ln: str) -> List[str]:
    Nsv = int(ln[29:32])  # Number of visible satellites this time %i3
    # get first 12 SV ID's
    sv = _getSVlist(ln, min(12, Nsv), [])

    # any more SVs?
    n = Nsv-12
    while n > 0:
        sv = _getSVlist(f.readline(), min(12, n), sv)
        n -= 12

    if Nsv != len(sv):
        raise LookupError('satellite list read incorrectly')

    return sv


def _getSVlist(ln: str, N: int, sv: List[str]) -> List[str]:
    """ parse a line of text from RINEX2 SV list"""
    sv.extend([ln[32+i*3:35+i*3] for i in range(N)])

    return sv


def obstime2(fn: Path) -> xarray.DataArray:
    """
    read all times in RINEX2 OBS file
    """
    times = []
    with opener(fn) as f:
        # Capture header info
        header = obsheader2(f)
        Nobs = header['Nobs']

        for ln in f:
            time_epoch = _timeobs(ln)
            if time_epoch is None:
                continue

            times.append(time_epoch)

            _skip(f, ln, Nobs)

    timedat = xarray.DataArray(times,
                               dims=['time'],
                               attrs={'filename': fn.name,
                                      'interval': header['interval']})

    return timedat


def _skip(f: TextIO, ln: str, Nobs: int, sv: Sequence[str]=None):
    """
    skip ahead to next time step
    """
    if sv is None:
        sv = _getsvind(f, ln)

    Nl_sv = ceil(Nobs/5)

    # f.seek(len(sv)*Nl_sv*80, 1)  # not for io.TextIOWrapper ?
    for _ in range(len(sv)*Nl_sv):
        f.readline()


def _timeobs(ln: str) -> Optional[datetime]:
    """
    Python >= 3.7 supports nanoseconds.  https://www.python.org/dev/peps/pep-0564/
    Python < 3.7 supports microseconds.
    """
    try:
        year = int(ln[1:3])
        if year < 80:
            year += 2000
        else:
            year += 1900

        return datetime(year=year,
                        month=int(ln[4:6]),
                        day=int(ln[7:9]),
                        hour=int(ln[10:12]),
                        minute=int(ln[13:15]),
                        second=int(ln[16:18]),
                        microsecond=int(float(ln[16:26]) % 1 * 1000000)
                        )
    except ValueError:  # garbage between header and RINEX data
        logging.warning(f'garbage detected in RINEX file')
        return None
